#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from unittest.mock import patch

import pytest

from airflow.models import DAG, DagRun, TaskInstance
from airflow.models.serialized_dag import SerializedDagModel
from airflow.providers.redis.log.redis_task_handler import RedisTaskHandler
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.utils.session import create_session
from airflow.utils.state import State
from airflow.utils.timezone import datetime

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.file_task_handler import extract_events
from tests_common.test_utils.version_compat import (
    AIRFLOW_V_3_0_PLUS,
    get_base_airflow_version_tuple,
)


class TestRedisTaskHandler:
    @pytest.fixture
    def ti(self):
        date = datetime(2020, 1, 1)
        dag = DAG(dag_id="dag_for_testing_redis_task_handler", schedule=None, start_date=date)
        task = EmptyOperator(task_id="task_for_testing_redis_log_handler", dag=dag)
        if AIRFLOW_V_3_0_PLUS:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                logical_date=date,
                data_interval=(date, date),
                run_after=date,
                run_id="test",
                run_type="scheduled",
            )
        else:
            dag_run = DagRun(
                dag_id=dag.dag_id,
                execution_date=date,
                run_id="test",
                run_type="scheduled",
            )

        dag_run.set_state(State.RUNNING)
        with create_session() as session:
            session.add(dag_run)
            session.commit()
            session.refresh(dag_run)

        if AIRFLOW_V_3_0_PLUS:
            from airflow.models.dag_version import DagVersion

            dag.sync_to_db()
            SerializedDagModel.write_dag(dag, bundle_name="testing")
            dag_version = DagVersion.get_latest_version(dag.dag_id)
            ti = TaskInstance(task=task, run_id=dag_run.run_id, dag_version_id=dag_version.id)
        else:
            ti = TaskInstance(task=task, run_id=dag_run.run_id)
        ti.dag_run = dag_run
        ti.try_number = 1
        ti.state = State.RUNNING

        yield ti

        with create_session() as session:
            session.query(DagRun).delete()

    @pytest.mark.db_test
    @conf_vars({("logging", "remote_log_conn_id"): "redis_default"})
    def test_write(self, ti):
        handler = RedisTaskHandler("any", max_lines=5, ttl_seconds=2)
        handler.set_context(ti)
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        key = (
            "dag_id=dag_for_testing_redis_task_handler/run_id=test"
            "/task_id=task_for_testing_redis_log_handler/attempt=1.log"
        )

        with patch("redis.Redis.pipeline") as pipeline:
            logger.info("Test log event")

        pipeline.return_value.rpush.assert_called_once_with(key, "Test log event")
        pipeline.return_value.ltrim.assert_called_once_with(key, start=-5, end=-1)
        pipeline.return_value.expire.assert_called_once_with(key, time=2)
        pipeline.return_value.execute.assert_called_once_with()

    @pytest.mark.db_test
    @conf_vars({("logging", "remote_log_conn_id"): "redis_default"})
    def test_read(self, ti):
        handler = RedisTaskHandler("any")
        handler.set_context(ti)
        logger = logging.getLogger(__name__)
        logger.addHandler(handler)

        key = (
            "dag_id=dag_for_testing_redis_task_handler/run_id=test"
            "/task_id=task_for_testing_redis_log_handler/attempt=1.log"
        )

        with patch("redis.Redis.lrange") as lrange:
            lrange.return_value = [b"Line 1", b"Line 2"]
            logs = handler.read(ti)

        if AIRFLOW_V_3_0_PLUS:
            if get_base_airflow_version_tuple() < (3, 1, 0):
                assert logs == (["Line 1\nLine 2"], {"end_of_log": True})
            else:
                log_stream, metadata = logs
                assert extract_events(log_stream) == ["Line 1", "Line 2"]
                assert metadata == {"end_of_log": True}
        else:
            assert logs == ([[("", "Line 1\nLine 2")]], [{"end_of_log": True}])
        lrange.assert_called_once_with(key, start=0, end=-1)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.models.dag import DAG
from airflow.providers.redis.hooks.redis import RedisHook
from airflow.providers.redis.sensors.redis_key import RedisKeySensor
from airflow.utils import timezone

DEFAULT_DATE = timezone.datetime(2017, 1, 1)


@pytest.mark.integration("redis")
class TestRedisSensor:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}

        self.dag = DAG("test_dag_id", default_args=args)
        self.sensor = RedisKeySensor(
            task_id="test_task", redis_conn_id="redis_default", dag=self.dag, key="test_key"
        )

    def test_poke(self):
        hook = RedisHook(redis_conn_id="redis_default")
        redis = hook.get_conn()
        redis.set("test_key", "test_value")
        assert self.sensor.poke(None), "Key exists on first call."
        redis.delete("test_key")
        assert not self.sensor.poke(None), "Key does NOT exists on second call."



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

from airflow.models.dag import DAG
from airflow.providers.redis.operators.redis_publish import RedisPublishOperator
from airflow.utils import timezone

DEFAULT_DATE = timezone.datetime(2017, 1, 1)


class TestRedisPublishOperator:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}

        self.dag = DAG("test_dag_id", schedule=None, default_args=args)

        self.mock_context = MagicMock()

    @patch("airflow.providers.redis.hooks.redis.RedisHook.get_conn")
    def test_execute_operator(self, mock_redis_conn):
        operator = RedisPublishOperator(
            task_id="test_task",
            dag=self.dag,
            channel="test_channel",
            message="test_message",
            redis_conn_id="redis_default",
        )
        operator.execute(self.mock_context)

        mock_redis_conn.assert_called_once_with()
        mock_redis_conn().publish.assert_called_once_with(channel="test_channel", message="test_message")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.redis.hooks.redis import RedisHook
from airflow.providers.redis.version_compat import BaseSensorOperator

if TYPE_CHECKING:
    try:
        from airflow.sdk.definitions.context import Context
    except ImportError:
        # TODO: Remove once provider drops support for Airflow 2
        from airflow.utils.context import Context


class RedisKeySensor(BaseSensorOperator):
    """Checks for the existence of a key in a Redis."""

    template_fields: Sequence[str] = ("key",)
    ui_color = "#f0eee4"

    def __init__(self, *, key: str, redis_conn_id: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.redis_conn_id = redis_conn_id
        self.key = key

    def poke(self, context: Context) -> bool:
        self.log.info("Sensor checks for existence of key: %s", self.key)
        return RedisHook(self.redis_conn_id).get_conn().exists(self.key)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from unittest.mock import patch

import pytest

from airflow.providers.redis.triggers.redis_await_message import AwaitMessageTrigger


class TestAwaitMessageTrigger:
    def test_trigger_serialization(self):
        trigger = AwaitMessageTrigger(
            channels=["test_channel"],
            redis_conn_id="redis_default",
            poll_interval=30,
        )

        assert isinstance(trigger, AwaitMessageTrigger)

        classpath, kwargs = trigger.serialize()

        assert classpath == "airflow.providers.redis.triggers.redis_await_message.AwaitMessageTrigger"
        assert kwargs == dict(
            channels=["test_channel"],
            redis_conn_id="redis_default",
            poll_interval=30,
        )

    @patch("airflow.providers.redis.hooks.redis.RedisHook.get_conn")
    @pytest.mark.asyncio
    async def test_trigger_run_succeed(self, mock_redis_conn):
        trigger = AwaitMessageTrigger(
            channels="test",
            redis_conn_id="redis_default",
            poll_interval=0.0001,
        )

        mock_redis_conn().pubsub().get_message.return_value = {
            "type": "message",
            "channel": "test",
            "data": "d1",
        }

        trigger_gen = trigger.run()
        task = asyncio.create_task(trigger_gen.__anext__())
        event = await task
        assert task.done() is True
        assert event.payload["data"] == "d1"
        assert event.payload["channel"] == "test"
        asyncio.get_event_loop().stop()

    @patch("airflow.providers.redis.hooks.redis.RedisHook.get_conn")
    @pytest.mark.asyncio
    async def test_trigger_run_fail(self, mock_redis_conn):
        trigger = AwaitMessageTrigger(
            channels="test",
            redis_conn_id="redis_default",
            poll_interval=0.01,
        )

        mock_redis_conn().pubsub().get_message.return_value = {
            "type": "subscribe",
            "channel": "test",
            "data": "d1",
        }

        trigger_gen = trigger.run()
        task = asyncio.create_task(trigger_gen.__anext__())
        await asyncio.sleep(1.0)
        assert task.done() is False
        task.cancel()
        asyncio.get_event_loop().stop()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from typing import Any

from asgiref.sync import sync_to_async

from airflow.providers.redis.hooks.redis import RedisHook
from airflow.triggers.base import BaseTrigger, TriggerEvent


class AwaitMessageTrigger(BaseTrigger):
    """
    A trigger that waits for a message matching specific criteria to arrive in Redis.

    The behavior of this trigger is as follows:
    - poll the Redis pubsub for a message, if no message returned, sleep

    :param channels: The channels that should be searched for messages
    :param redis_conn_id: The connection object to use, defaults to "redis_default"
    :param poll_interval: How long the trigger should sleep after reaching the end of the Redis log
        (seconds), defaults to 60
    """

    def __init__(
        self,
        channels: list[str] | str,
        redis_conn_id: str = "redis_default",
        poll_interval: float = 60,
    ) -> None:
        self.channels = channels
        self.redis_conn_id = redis_conn_id
        self.poll_interval = poll_interval

    def serialize(self) -> tuple[str, dict[str, Any]]:
        return (
            "airflow.providers.redis.triggers.redis_await_message.AwaitMessageTrigger",
            {
                "channels": self.channels,
                "redis_conn_id": self.redis_conn_id,
                "poll_interval": self.poll_interval,
            },
        )

    async def run(self):
        hook = RedisHook(redis_conn_id=self.redis_conn_id).get_conn().pubsub()
        hook.subscribe(self.channels)

        async_get_message = sync_to_async(hook.get_message)
        while True:
            message = await async_get_message()

            if message and message["type"] == "message":
                yield TriggerEvent(message)
                break
            await asyncio.sleep(self.poll_interval)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This is an example DAG which uses RedisPublishOperator, RedisPubSubSensor and RedisKeySensor.
In this example, we create 3 tasks which execute sequentially.
The first task is to publish a particular message to redis using the RedisPublishOperator.
The second task is to wait for a particular message at a particular channel to arrive in redis
using the RedisPubSubSensor, and the third task is to wait for a particular key to arrive in
redis using the RedisKeySensor.

"""

from __future__ import annotations

import os
from datetime import datetime

# [START import_module]
from airflow import DAG
from airflow.providers.redis.operators.redis_publish import RedisPublishOperator
from airflow.providers.redis.sensors.redis_key import RedisKeySensor

# [END import_module]
# [START instantiate_dag]
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID")

default_args = {
    "start_date": datetime(2023, 5, 15),
    "max_active_runs": 1,
}

with DAG(
    dag_id="redis_example",
    schedule=None,
    default_args=default_args,
) as dag:
    # [START RedisPublishOperator_DAG]
    publish_task = RedisPublishOperator(
        task_id="publish_task",
        redis_conn_id="redis_default",
        channel="your_channel",
        message="Start processing",
        dag=dag,
    )

    # [END RedisPublishOperator_DAG]

    # [START RedisKeySensor_DAG]
    key_sensor_task = RedisKeySensor(
        task_id="key_sensor_task",
        redis_conn_id="redis_default",
        key="your_key",
        dag=dag,
        timeout=600,
        poke_interval=30,
    )
    # [END RedisKeySensor_DAG]

    publish_task >> key_sensor_task

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, call, patch

from airflow.models.dag import DAG
from airflow.providers.redis.sensors.redis_pub_sub import RedisPubSubSensor
from airflow.utils import timezone

DEFAULT_DATE = timezone.datetime(2017, 1, 1)


class TestRedisPubSubSensor:
    def setup_method(self):
        args = {"owner": "airflow", "start_date": DEFAULT_DATE}

        self.dag = DAG("test_dag_id", schedule=None, default_args=args)

        self.mock_context = MagicMock()

    @patch("airflow.providers.redis.hooks.redis.RedisHook.get_conn")
    def test_poke_mock_true(self, mock_redis_conn):
        sensor = RedisPubSubSensor(
            task_id="test_task", dag=self.dag, channels="test", redis_conn_id="redis_default"
        )

        mock_redis_conn().pubsub().get_message.return_value = {
            "type": "message",
            "channel": b"test",
            "data": b"d1",
        }

        result = sensor.poke(self.mock_context)
        assert result

        context_calls = [
            call.xcom_push(key="message", value={"type": "message", "channel": b"test", "data": b"d1"})
        ]

        assert self.mock_context["ti"].method_calls == context_calls, "context call  should be same"

    @patch("airflow.providers.redis.hooks.redis.RedisHook.get_conn")
    def test_poke_mock_false(self, mock_redis_conn):
        sensor = RedisPubSubSensor(
            task_id="test_task", dag=self.dag, channels="test", redis_conn_id="redis_default"
        )

        mock_redis_conn().pubsub().get_message.return_value = {
            "type": "subscribe",
            "channel": b"test",
            "data": b"d1",
        }

        result = sensor.poke(self.mock_context)
        assert not result

        context_calls = []
        assert self.mock_context["ti"].method_calls == context_calls, "context calls should be same"

    @patch("airflow.providers.redis.hooks.redis.RedisHook.get_conn")
    def test_poke_subscribe_called_only_once(self, mock_redis_conn):
        sensor = RedisPubSubSensor(
            task_id="test_task", dag=self.dag, channels="test", redis_conn_id="redis_default"
        )

        mock_redis_conn().pubsub().get_message.return_value = {
            "type": "subscribe",
            "channel": b"test",
            "data": b"d1",
        }

        result = sensor.poke(self.mock_context)
        assert not result

        context_calls = []
        assert self.mock_context["ti"].method_calls == context_calls, "context calls should be same"
        result = sensor.poke(self.mock_context)

        assert mock_redis_conn().pubsub().subscribe.call_count == 1



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.models import Connection
from airflow.providers.redis.hooks.redis import RedisHook


class TestRedisHook:
    @pytest.mark.db_test
    def test_get_conn(self):
        hook = RedisHook(redis_conn_id="redis_default")
        assert hook.redis is None

        assert hook.host is None, "host initialised as None."
        assert hook.port is None, "port initialised as None."
        assert hook.password is None, "password initialised as None."
        assert hook.db is None, "db initialised as None."
        assert hook.get_conn() is hook.get_conn(), "Connection initialized only if None."

    @mock.patch("airflow.providers.redis.hooks.redis.Redis")
    @mock.patch(
        "airflow.providers.redis.hooks.redis.RedisHook.get_connection",
    )
    def test_get_conn_with_extra_config(self, mock_get_connection, mock_redis):
        connection = Connection(
            login="user",
            password="password",
            host="remote_host",
            port=1234,
        )
        connection.set_extra(
            """{
                "db": 2,
                "ssl": true,
                "ssl_cert_reqs": "required",
                "ssl_ca_certs": "/path/to/custom/ca-cert",
                "ssl_keyfile": "/path/to/key-file",
                "ssl_certfile": "/path/to/cert-file",
                "ssl_check_hostname": true
            }"""
        )
        mock_get_connection.return_value = connection
        hook = RedisHook()

        hook.get_conn()
        mock_redis.assert_called_once_with(
            host=connection.host,
            username=connection.login,
            password=connection.password,
            port=connection.port,
            db=connection.extra_dejson["db"],
            ssl=connection.extra_dejson["ssl"],
            ssl_cert_reqs=connection.extra_dejson["ssl_cert_reqs"],
            ssl_ca_certs=connection.extra_dejson["ssl_ca_certs"],
            ssl_keyfile=connection.extra_dejson["ssl_keyfile"],
            ssl_certfile=connection.extra_dejson["ssl_certfile"],
            ssl_check_hostname=connection.extra_dejson["ssl_check_hostname"],
        )

    @pytest.mark.db_test
    def test_get_conn_password_stays_none(self):
        hook = RedisHook(redis_conn_id="redis_default")
        hook.get_conn()
        assert hook.password is None
